# -*- coding: utf-8 -*-

"""Top-level package for hasasia."""

__author__ = """Jeffrey S. Hazboun"""
__email__ = 'jeffrey.hazboun@gmail.com'
__version__ = '0.1.4'

citation = 'Hazboun, J.S., Romano, J.R., Smith, T.L. '
citation += 'Realistic sensitivity curves for pulsar timing arrays. '
citation += 'arxiv:1907.04341'
__cite__ = citation
